#include "DHT.h"
#include <LiquidCrystal.h>

#define DHTPIN 13
#define DHTTYPE DHT22
DHT dht(DHTPIN, DHTTYPE);

#define GAS A2

#define AIR_QUALITY A1

#define BUZZER 12

#define LIGHT A3

#define MOTION A4

#define SOUND A5

#define RGB_R 11
#define RGB_G 3
#define RGB_B 2

typedef struct roomstate {
  float temperature, humidity, heatIndex;
  int airQuality;
  int airQualityPercentage;
  int lightIntensity;
  int gasValue;
  int motion;
  int sound;
} RoomState;

RoomState roomState;
String hour, date, weekDay;

int tick = 0;
int switchLCD = -1;
int sleepMode = 0;

const int pin_RS = 8;
const int pin_EN = 9;
const int pin_d4 = 4;
const int pin_d5 = 5;
const int pin_d6 = 6;
const int pin_d7 = 7;
const int pin_BL = 10;
LiquidCrystal lcd( pin_RS,  pin_EN,  pin_d4,  pin_d5,  pin_d6,  pin_d7);

void setup() {
  Serial.begin(9600);
  Serial.setTimeout(150);

  dht.begin();
  pinMode(10, OUTPUT);

  pinMode(BUZZER, OUTPUT);
  pinMode(MOTION, INPUT);
  pinMode(GAS, INPUT);
  pinMode(AIR_QUALITY, INPUT);
  pinMode(LIGHT, INPUT);
  pinMode(SOUND, INPUT);

  pinMode(RGB_R, OUTPUT);
  pinMode(RGB_G, OUTPUT);
  pinMode(RGB_B, OUTPUT);

  lcd.begin(16, 2);
}


void getHour()
{
  Serial.println("getHour");
  hour = Serial.readString();
}

void getDate()
{
  Serial.println("getDate");
  date = Serial.readString();
}

void getWeekday()
{
  Serial.println("getWeekday");
  weekDay = Serial.readString();
}

void buzz(float seconds, int sign)
{
  tone(BUZZER, sign);
  delay(seconds * 1000);
  noTone(BUZZER);
}

void RGB_color(int red_light_value, int green_light_value, int blue_light_value)
{
  analogWrite(RGB_R, red_light_value);
  analogWrite(RGB_G, green_light_value);
  analogWrite(RGB_B, blue_light_value);
}

int handleGasSensor()
{
  roomState.gasValue = analogRead(GAS);

  if (roomState.gasValue > 100)
  {
    lcd.clear();
    RGB_color(255, 0, 0);
    lcd.print("  GAS DETECTED ");
    lcd.setCursor(0, 1);
    lcd.print("OPEN THE WINDOWS");
    buzz(0.5, 1000);
    RGB_color(0, 0, 0);
    delay(100);
    lcd.clear();
    return 1;
  }

  return 0;
}

void handleLED()
{

  if (roomState.airQualityPercentage < 50)
    RGB_color(255, 0, 0);
  else
    RGB_color(0, 255, 0);
}

void handleAirQualitySensor()
{
  roomState.airQuality = analogRead(AIR_QUALITY);

  roomState.airQualityPercentage = 100 - map(roomState.airQuality, 0, 500, 0, 100);
}

void handleLightSensor()
{
  roomState.lightIntensity = digitalRead(LIGHT);
}

void handleMotionSensor()
{
  roomState.motion = digitalRead(MOTION);
}

int handleTemperatureSensor()
{
  float h = dht.readHumidity();
  float t = dht.readTemperature();

  if (isnan(h) || isnan(t)) {
    Serial.println(F("Failed to read from DHT sensor!"));
    return -1;
  }

  float hic = dht.computeHeatIndex(t, h, false);

  roomState.humidity = h;
  roomState.temperature = t;
  roomState.heatIndex = hic;
  return 0;
}

void handleSoundSensor()
{
  int sound = 0;

  for (int i = 0; i < 100; ++i)
    sound += analogRead(SOUND);

  roomState.sound = sound / 100;
}

void printLCDTemperature()
{
  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print("Temp: " + String(roomState.temperature) + String("C") + String(" ") + String((int)roomState.humidity) + String("%"));
  lcd.setCursor(0, 1);
  lcd.print("RealFeel: " + String(roomState.heatIndex) + String("C"));
}

void printLCDAirQuality()
{
  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print("Air quality: " + String(roomState.airQualityPercentage) + String ("%"));
  lcd.setCursor(0, 1);

  if (roomState.heatIndex < 26)
    lcd.print("HeatIndex: GOOD");
  else if (roomState.heatIndex >= 26 && roomState.heatIndex < 32)
    lcd.print("HeatIndex:   BAD");
  else
    lcd.print("HeatIndex:  RISK");
}

void printLCDDateTime()
{
  getHour();
  getDate();
  getWeekday();

  lcd.clear();

  if (hour == NULL || date == NULL)
  {
    lcd.setCursor(0, 0);
    lcd.print(" Error fetching ");
    lcd.setCursor(0, 1);
    lcd.print(" the date/time! ");
  }
  else
  {
    lcd.setCursor(0, 0);
    lcd.print(String("Hour: ") + hour);
    lcd.setCursor(0, 1);
    lcd.print(String("Date: ") + date);
    lcd.setCursor(12, 0);
    lcd.print(weekDay);
  }
}

void handleLCDScreen()
{
  if (tick % 5 == 0)
  {
    handleLED();

    switchLCD++;

    if (switchLCD == 3)
      switchLCD = 0;

    if (switchLCD == 0)
      printLCDTemperature();
    else if (switchLCD == 1)
      printLCDAirQuality();
    else if (switchLCD == 2)
      printLCDDateTime();

    tick = 0;
  }

  tick++;
}

void flushRoomState()
{
  Serial.println("flush");

  Serial.println(roomState.temperature);
  Serial.println(roomState.humidity);
  Serial.println(roomState.heatIndex);
  Serial.println(roomState.airQuality);
  Serial.println(roomState.airQualityPercentage);
  Serial.println(roomState.lightIntensity);
  Serial.println(roomState.gasValue);
  Serial.println(roomState.motion);
  Serial.println(roomState.sound);

}

int gasDetected = 0;

void handleSensors()
{
  handleAirQualitySensor();
  handleLightSensor();
  handleTemperatureSensor();
  handleLightSensor();
  handleMotionSensor();
  handleSoundSensor();
}


void clearLCDLine(int line)
{
  lcd.setCursor(0, line);
  for (int n = 0; n < 20; n++) // 20 indicates symbols in line. For 2x16 LCD write - 16
  {
    lcd.print(" ");
  }
}

void initSensors
{
  digitalWrite(10, HIGH);
  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print("Warming sensors!");

  for (int i = 120; i >= 0; i--)
  {

    lcd.setCursor(0, 1);

    if (i == 120)
      lcd.print(String("     ") + String("02:00") + String("     "));
    else if (i < 120 && i >= 60)
    {
      if (i < 70)
        lcd.print(String("     ") + String("01:0") + String(i - 60) + String("     "));
      else
        lcd.print(String("     ") + String("01:") + String(i - 60) + String("     "));
    }
    else
    {
      if (i < 10)
        lcd.print(String("     ") + String("00:0") + String(i) + String("     "));
      else
        lcd.print(String("     ") + String("00:") + String(i) + String("     "));
    }

    delay(1100);
  }
}


int Initialize = 1;
int flushCounter = 0;

void loop() {

  if (Initialize)
  {
    Initialize = 0;
    initSensors();
    handleSensors();
  }

 
  flushCounter++;
  gasDetected = handleGasSensor();
  handleMotionSensor();

  if (roomState.motion == LOW && sleepMode == 0)
  {
    sleepMode = 1;
    RGB_color(0, 0, 0);
    lcd.clear();
    digitalWrite(10, LOW);
    switchLCD = -1;
    tick = 0;
  }
  else if (roomState.motion == HIGH) {
    sleepMode = 0;
    digitalWrite(10, HIGH);
    if (!gasDetected)
      handleLCDScreen();
  }

  delay(1000);
  
  if (flushCounter == 300)
  {
    handleSensors();
    flushRoomState();
    flushCounter = 0;
  }
}
